from typing import Type, Iterable, Dict

from antipickle.base import AbstractAdapter, wrap


class DataclassAdapter(AbstractAdapter):
    typestring = 'dataclass'

    def __init__(self, name2class: Dict[str, Type]):
        self.name2class = name2class
        self.class2name = {v: k for k, v in self.name2class.items()}

    def check_type(self, obj):
        # we intentionally ignore inherited classes
        return type(obj) in self.class2name

    def to_dict(self, obj):
        return {
            'classname': self.class2name[type(obj)],
            # we don't use asdict, because it doesn't reverse that simple
            'data': {k: wrap(v) for k, v in obj.__dict__.items()},
        }

    def from_dict(self, d):
        classobj = self.name2class[d['classname']]
        return classobj(**d['data'])


class PydanticAdapter(AbstractAdapter):
    typestring = 'pydantic'

    def __init__(self, name2pydantic_class: Dict[str, Type]):
        self.name2class = name2pydantic_class
        self.class2name = {v: k for k, v in self.name2class.items()}

    def check_type(self, obj):
        # we intentionally ignore inherited classes
        return type(obj) in self.class2name

    def to_dict(self, obj):
        return {
            'classname': self.class2name[type(obj)],
            'data': {k: wrap(v) for k, v in obj.dict().items()}
        }

    def from_dict(self, d):
        classobj = self.name2class[d['classname']]
        return classobj(**d['data'])


class SkipSerialization(AbstractAdapter):
    typestring = 'skipped'

    def __init__(self, skipped_types: Iterable[Type]):
        self.skipped_types = set(skipped_types)

    def check_type(self, obj):
        # we intentionally ignore inherited classes
        return type(obj) in self.skipped_types

    def to_dict(self, obj):
        return {
            'skipped': True,
        }

    def from_dict(self, d):
        return 'skipped, as specified during serialization'



import datetime
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional, List, Tuple

import numpy as np
import pandas as pd
import antipickle
from antipickle.base import MAGIC_KEY


def get_serializers():
    with TemporaryDirectory() as dir:
        fname = Path(dir).joinpath('file')
        fname_gz = Path(dir).joinpath('file.gz')

        def pack_to_file(o):
            antipickle.dump(o, fname)
            return fname

        def unpack_from_file(fname):
            return antipickle.load(fname)

        yield pack_to_file, unpack_from_file

        def pack_to_file_gz(o):
            antipickle.dump(o, fname_gz)
            return fname_gz

        yield pack_to_file_gz, unpack_from_file

        path = Path.home().joinpath('.test_paths_for_antipickle.antipickle')
        if path.exists():
            # use antipickle to test antipickle, what an irony
            paths = antipickle.load(path)

            def pack_to_s3(o):
                result_path = paths['s3_file_antipickle']
                antipickle.dump(o, result_path)
                return result_path

            yield pack_to_s3, unpack_from_file

            def pack_to_s3_gz(o):
                result_path = paths['s3_file_antipickle_gz']
                antipickle.dump(o, result_path)
                return result_path

            yield pack_to_s3_gz, unpack_from_file


def test_simple_types():
    now = datetime.datetime.now()

    x = dict(
        a=None,
        b='string',
        c=b'bytes',
        d=12345,
        e=123.456,
        f=['a', 'b', 'c', b'bytes', True, False],
        g={1, 2, 3},
        h=now,
        i=now.time(),
        j=now.date(),
        k=datetime.timedelta(weeks=3),
        l=(1 + 2.2j),
        m=(1, [2, 3], {4: 5, (6, 'string'): {None, now}}),
        # n=[10 ** 30, -10 ** 30] # fails for integers out of 64-bit range
    )

    for serializer, deserializer in get_serializers():
        assert x == deserializer(serializer(x))

        y = {**x, MAGIC_KEY: {MAGIC_KEY: (1, 2)}}
        y = {**y, '123': {**y, 1: (2, y)}}
        assert y == deserializer(serializer(y))

        z = [y]
        assert z == deserializer(serializer(z))

        z = tuple(z)
        assert z == deserializer(serializer(z))


def test_numpy():
    d = dict(
        a=np.zeros([1, 2, 3], dtype='uint8'),
        b=np.asarray(['hi'] * 8),
        c=np.arange(100).astype('float32').reshape([2, 5, 10]),
        d=np.bool_(1),
        e=np.float32(15.),
        # f=np.float64(15.), # auto-converts to python built-in float
        g=np.float16(121.),
    )
    for serializer, deserializer in get_serializers():
        d2 = deserializer(serializer(d))
        for k in d:
            l, r = d[k], d2[k]
            assert type(l) == type(r), [type(l), type(r)]
            assert l.dtype == r.dtype
            assert np.array_equal(l, r)

        for special_number in [np.nan, np.infty, -np.infty]:
            result = deserializer(serializer(special_number))
            assert type(result) == type(special_number)
            assert np.isnan(result) == np.isnan(special_number)
            assert np.isinf(result) == np.isinf(special_number)


def test_pandas():
    df = pd.DataFrame(dict(a=[1, 2, 3], b=[True, False, False], c=['one', 'two', 'three'], d=[1., 2., 3.]))
    df['arrayed'] = list(np.arange(3 * 4).reshape([3, 4]))

    # three levels of columns
    df2 = df.copy()
    df2.columns = pd.MultiIndex.from_tuples([(x, x, x) for x in df.columns])

    d = dict(
        a=df,
        b=df.set_index(['a', 'b']),
        c=df2,
    )
    for serializer, deserializer in get_serializers():
        d2 = deserializer(serializer(d))
        for k in d:
            l, r = d[k], d2[k]
            assert type(l) == type(r), [type(l), type(r)]
            assert (l.dtypes == r.dtypes).all()
            assert pd.DataFrame.equals(l, r)

        cols = [df[c] for c in df.columns]
        for l, r in zip(
                cols,
                deserializer(serializer(cols)),
        ):
            assert type(l) == type(r)
            assert l.name == r.name


def test_polars():
    import polars as pl
    df = pl.DataFrame(dict(
        a=[1, 2, 3],
        b=[True, False, False],
        c=['one', 'two', 'three'],
        d=[1., 2., 3.],
        arrayed=[[1, 2], [3, 4], [5, 6]],
    ))

    for serializer, deserializer in get_serializers():
        df2 = deserializer(serializer(df))

        assert type(df) == type(df2), [type(df), type(df2)]
        # assert (l.dtypes == r.dtypes).all()
        assert df.frame_equal(df2)

        cols = [df.get_column(c) for c in df.columns]
        for l, r in zip(
            cols,
            deserializer(serializer(cols)),
        ):
            assert type(l) == type(r)
            assert l.name == r.name


def test_dataclasses():
    from dataclasses import dataclass

    @dataclass
    class Point:
        x: int
        y: float

    @dataclass
    class Point3d(Point):
        z: str
        w: Optional[np.ndarray] = None

    @dataclass
    class Polygon:
        points_2d: List[Point]
        point_3d: Point3d

    from antipickle.adapters import DataclassAdapter
    from antipickle import dumps, loads
    adapters = [DataclassAdapter(dict(
        point=Point,
        point_3d=Point3d,
        poly=Polygon,
    ))]

    l = Polygon(
        points_2d=[Point(1, 2.), Point(3, 4.)],
        point_3d=Point3d(3, 4., '5', w=np.zeros(1)),
    )

    packed = dumps(l, adapters=adapters)
    r = loads(packed, adapters=adapters)
    assert l == r


def test_pydantic():
    from pydantic import BaseModel

    class Point(BaseModel):
        x: int
        y: float

    class Point3d(Point):
        z: str
        w: Tuple[str, str] = None

    class Polygon(BaseModel):
        points_2d: List[Point]
        point_3d: Point3d

    from antipickle.adapters import PydanticAdapter
    from antipickle import dumps, loads
    adapters = [PydanticAdapter(dict(
        point=Point,
        point_3d=Point3d,
        poly=Polygon,
    ))]

    l = Polygon(
        points_2d=[Point(x=1, y=2.), Point(x=3, y=4.)],
        point_3d=Point3d(x=3, y=4., z='5', w=('a', 'bc')),
    )

    packed = dumps(l, adapters=adapters)
    r = loads(packed, adapters=adapters)
    assert l == r


if __name__ == '__main__':
    test_pydantic()
    test_dataclasses()
    test_simple_types()
    test_numpy()
    test_pandas()



"""
These adapters are turned on by default and require no customization
"""
import datetime as dt
from io import BytesIO
from typing import List

from antipickle.base import AbstractAdapter, wrap


class SetAdapter(AbstractAdapter):
    typestring = 'set'

    def check_type(self, obj):
        return type(obj) == set

    def to_dict(self, obj):
        return {'data': [wrap(x) for x in obj]}

    def from_dict(self, d):
        return set(d['data'])


class TupleAdapter(AbstractAdapter):
    typestring = 'tuple'

    def check_type(self, obj):
        return type(obj) == tuple

    def to_dict(self, obj):
        return {'data': [wrap(x) for x in obj]}

    def from_dict(self, d):
        return tuple(d['data'])


class NumpyAdapter(AbstractAdapter):
    typestring = 'np.ndarray'
    package_name = 'numpy'

    def check_type(self, obj):
        import numpy as np
        return isinstance(obj, np.ndarray)

    def to_dict(self, obj):
        import numpy as np
        bytes_io = BytesIO()
        np.save(bytes_io, obj, allow_pickle=False, fix_imports=False)
        return {'data': bytes_io.getvalue()}

    def from_dict(self, d):
        import numpy as np
        bytes_io = BytesIO(d['data'])
        return np.load(bytes_io, allow_pickle=False)


class NumpyScalarAdapter(AbstractAdapter):
    typestring = 'np.scalar'
    package_name = 'numpy'

    def check_type(self, obj):
        import numpy as np
        return isinstance(obj, (np.bool_, np.number))

    def to_dict(self, obj):
        import numpy as np
        bytes_io = BytesIO()
        np.save(bytes_io, np.asarray(obj), allow_pickle=False, fix_imports=False)
        return {'data': bytes_io.getvalue()}

    def from_dict(self, d):
        import numpy as np
        bytes_io = BytesIO(d['data'])
        arr = np.load(bytes_io, allow_pickle=False)
        assert arr.ndim == 0
        return arr[()]


class DateTimeAdapter(AbstractAdapter):
    typestring = 'dt.datetime'

    def check_type(self, obj):
        return isinstance(obj, dt.datetime)

    def to_dict(self, obj):
        return {'data': dt.datetime.isoformat(obj)}

    def from_dict(self, d):
        return dt.datetime.fromisoformat(d['data'])


class DateAdapter(AbstractAdapter):
    typestring = 'dt.date'

    def check_type(self, obj):
        return isinstance(obj, dt.date)

    def to_dict(self, obj):
        return {'data': dt.date.isoformat(obj)}

    def from_dict(self, d):
        return dt.date.fromisoformat(d['data'])


class TimeAdapter(AbstractAdapter):
    typestring = 'dt.time'

    def check_type(self, obj):
        return isinstance(obj, dt.time)

    def to_dict(self, obj):
        return {'data': dt.time.isoformat(obj)}

    def from_dict(self, d):
        return dt.time.fromisoformat(d['data'])


class TimedeltaAdapter(AbstractAdapter):
    typestring = 'dt.timedelta'

    def check_type(self, obj):
        return isinstance(obj, dt.timedelta)

    def to_dict(self, obj):
        return {'data': dt.timedelta.total_seconds(obj)}

    def from_dict(self, d):
        return dt.timedelta(seconds=d['data'])


class ComplexAdapter(AbstractAdapter):
    typestring = 'complex'

    def check_type(self, obj):
        return isinstance(obj, complex)

    def to_dict(self, obj):
        return {'data': complex.__repr__(obj)}

    def from_dict(self, d):
        return complex(d['data'])


class PandasDataFrameAdapter(AbstractAdapter):
    typestring = 'pd.DataFrame'
    package_name = 'pandas'

    def check_type(self, obj):
        import pandas as pd
        return isinstance(obj, pd.DataFrame)

    def to_dict(self, obj):
        bytes_io = BytesIO()
        obj.to_parquet(bytes_io, engine='pyarrow')
        return {'data': bytes_io.getvalue()}

    def from_dict(self, d):
        import pandas as pd
        return pd.read_parquet(BytesIO(d['data']))


class PandasSeriesAdapter(AbstractAdapter):
    typestring = 'pd.Series'
    package_name = 'pandas'

    def check_type(self, obj):
        import pandas as pd
        return isinstance(obj, pd.Series)

    def to_dict(self, obj):
        # series is converted to DF, then serialized with parquet
        bytes_io = BytesIO()
        obj.to_frame().to_parquet(bytes_io, engine='pyarrow')
        return {'data': bytes_io.getvalue()}

    def from_dict(self, d):
        # read dataframe, then take the first column
        import pandas as pd
        df = pd.read_parquet(BytesIO(d['data']))
        return df.iloc[:, 0]


class PolarsDataFrameAdapter(AbstractAdapter):
    typestring = 'pl.DataFrame'
    package_name = 'polars'

    def check_type(self, obj):
        import polars as pl
        return isinstance(obj, pl.DataFrame)

    def to_dict(self, obj):
        bytes_io = BytesIO()
        obj.write_parquet(bytes_io)
        return {'data': bytes_io.getvalue()}

    def from_dict(self, d):
        import polars as pl
        return pl.read_parquet(BytesIO(d['data']))


class PolarsSeriesAdapter(AbstractAdapter):
    typestring = 'pl.Series'
    package_name = 'polars'

    def check_type(self, obj):
        import polars as pl
        return isinstance(obj, pl.Series)

    def to_dict(self, obj):
        bytes_io = BytesIO()
        # convert to frame and write to parquet
        obj.to_frame().write_parquet(bytes_io)
        return {'data': bytes_io.getvalue()}

    def from_dict(self, d):
        import polars as pl
        df = pl.read_parquet(BytesIO(d['data']))
        # read parquet and take the first column of dataframe
        assert df.shape[1] == 1, 'more than one column in df. Sounds wrong'
        return df[:, 0]

# deserialization in scipy.sparse still produces matrices, not arrays :(
# there is no point in supporting matrices at this point.

# class ScipySparseAdapter(AbstractAdapter):
#     typestring = 'sc.sparse'
#     package_name = 'scipy.sparse'
#
#     def check_type(self, obj):
#         from scipy import sparse
#         # process only arrays, not matrices as those will be dropped
#         return isinstance(obj, (
#             sparse.csc_array,
#             sparse.csr_array,
#             sparse.coo_array,
#             sparse.lil_array,
#             sparse.dia_array,
#             sparse.dok_array,
#         ))
#
#     def to_dict(self, obj):
#         from scipy.sparse import save_npz
#         b = BytesIO()
#         save_npz(b, obj, compressed=True)
#         return {'data': b.getvalue()}
#
#     def from_dict(self, d):
#         from scipy.sparse import load_npz
#         return load_npz(BytesIO(d['data']))


all_default_adapters: List[AbstractAdapter] = [
    SetAdapter(),
    TupleAdapter(),
    NumpyAdapter(),
    NumpyScalarAdapter(),
    ComplexAdapter(),
    TimeAdapter(),
    TimedeltaAdapter(),
    DateTimeAdapter(),
    DateAdapter(),
    PandasSeriesAdapter(),
    PandasDataFrameAdapter(),
    PolarsDataFrameAdapter(),
    PolarsSeriesAdapter(),
]



from __future__ import annotations

import contextlib
import gzip
import sys
from io import BytesIO
from pathlib import Path
from typing import Optional, Any, List, IO, Iterable, Union

import msgpack

from .base import wrap, TypeMatcher, DictAdapter, AbstractAdapter

__version__ = '0.2.0pre'
format_version: int = 1

# Should never be changed
INTRODUCTION_PHRASE = 'unpack me with antipickle'

# required by .isoformat in python's datetime
assert sys.version_info >= (3, 7), 'Old python is not supported by antipickle'


class AntipickleSerializationError(RuntimeError):
    pass


class AntipickleDeserializationError(RuntimeError):
    pass


def get_matcher(adapters: Optional[list]) -> TypeMatcher:
    from .adapters_default import all_default_adapters
    matcher = TypeMatcher()
    matcher.register_adapter(DictAdapter())
    if adapters is not None:
        for c in adapters:
            matcher.register_adapter(c)
    for c in all_default_adapters:
        matcher.register_adapter(c)
    return matcher


class Packer:
    def __init__(self, f: IO[bytes], *, adapters: Optional[List[AbstractAdapter]] = None):
        self.f = f
        self.matcher = get_matcher(adapters)
        metadata = None
        self.f.write(
            msgpack.packb([INTRODUCTION_PHRASE, format_version, metadata])
        )

    def add(self, obj: Any):
        self.f.write(
            msgpack.packb(wrap(obj), default=self.matcher.serialize_helper)
        )


def load_sequence(f: IO[bytes], *, adapters: Optional[List[AbstractAdapter]] = None) -> Iterable[Any]:
    matcher = get_matcher(adapters=adapters)
    unpacker = msgpack.Unpacker(f, object_hook=matcher.deserialize_helper)
    intro = next(unpacker)
    if not isinstance(intro, list) \
            or len(intro) != 3 \
            or intro[0] != INTRODUCTION_PHRASE \
            or not isinstance(intro[1], int):
        raise AntipickleDeserializationError('Wrong format or corrupted file')
    if intro[1] > format_version:
        raise AntipickleDeserializationError(f'Antipickle format {format_version} not supported; update antipickle')
    yield from unpacker


def dump_sequence(sequence: Iterable[Any], f: IO[bytes],
                  *, adapters: Optional[List[AbstractAdapter]] = None):
    packer = Packer(f, adapters=adapters)
    for x in sequence:
        packer.add(x)


@contextlib.contextmanager
def get_file_handle(filename: str, mode: str):
    assert mode in ('rb', 'wb')
    use_fsspec = "://" in filename
    if not use_fsspec:
        handle = open(filename, mode=mode)
    else:
        import fsspec
        handle = fsspec.open(filename, mode=mode)

    if filename.endswith('.gz'):
        with handle as fraw:
            with gzip.open(fraw, mode=mode) as f:
                yield f
    else:
        with handle as f:
            yield f


def dump(
        obj: Any,
        filename: Union[str, Path],
        *, adapters: Optional[List[AbstractAdapter]] = None
):
    assert isinstance(filename, (str, Path)), f'Passed filename is of a wrong type: {type(filename)}'
    filename = str(filename)
    assert len(filename) < 1000, 'Filename is too long, check order of arguments'
    with get_file_handle(filename, mode='wb') as f:
        dump_sequence([obj], f, adapters=adapters)


def load(
        filename: Union[str, Path],
        *, adapters: Optional[List[AbstractAdapter]] = None
) -> Any:
    assert isinstance(filename, (str, Path)), f'Passed filename is of a wrong type {type(filename)}'
    filename = str(filename)
    with get_file_handle(filename, mode='rb') as f:
        loaded = list(load_sequence(f, adapters=adapters))

    if len(loaded) == 0:
        raise AntipickleDeserializationError('File is empty')
    if len(loaded) != 1:
        raise AntipickleDeserializationError('More than one object is stored, use antipickle.load_sequence')
    return loaded[0]


def dumps(obj: Any, *, adapters: Optional[List[AbstractAdapter]] = None) -> bytes:
    b = BytesIO()
    dump_sequence([obj], f=b, adapters=adapters)
    return b.getvalue()


def loads(b: bytes, *, adapters: Optional[List[AbstractAdapter]] = None) -> Any:
    loaded = list(load_sequence(BytesIO(b), adapters=adapters))
    if len(loaded) == 0:
        raise AntipickleDeserializationError('File is empty')
    if len(loaded) != 1:
        raise AntipickleDeserializationError('More than one object is stored, use antipickle.load_sequence')
    return loaded[0]



"""
Main classes/functions used for conversion.
"""
from __future__ import annotations
import dataclasses
import sys
import typing
from typing import Any, Optional, Dict

# special key in dictionary.
# Its presence means it is a custom type and should be processed by package
# It was selected because of being a rarely used ASCII symbol.
MAGIC_KEY = '\x7f'

"""
We use wrapper (LazyWrapper) to wrap every data component 
except for the most trivial ones (ints, floats, str, ...)

Thus each time serializer meets wrapper, it asks how the object should be processed.
We provide such hooks for serializers / deserializers.
Hook scans for available (de)serializers for this type and applies the right one.
"""

NoneType: Any = type(None)


def wrap(obj):
    if isinstance(obj, (str, float, int, bool, bytes, NoneType)):
        return obj

    if isinstance(obj, list):
        return [wrap(x) for x in obj]
    return LazyWrapper(obj=obj)


@dataclasses.dataclass
class LazyWrapper:
    obj: typing.Any


class TypeMatcher:
    def __init__(self):
        self.adapters: Dict[str, AbstractAdapter] = {}
        self.type2adapter: Dict[Any, AbstractAdapter] = {}
        self.serialization_stack = set()

    def register_adapter(self, adapter: AbstractAdapter):
        assert isinstance(adapter, AbstractAdapter)
        assert adapter.typestring not in self.adapters
        self.adapters[adapter.typestring] = adapter

    def _serialize_object(self, x):
        t = type(x)
        x_id = id(x)
        assert x_id not in self.serialization_stack
        self.serialization_stack.add(x_id)

        if type(x) not in self.type2adapter:
            for adapter in self.adapters.values():
                if adapter.package_name is not None and adapter.package_name not in sys.modules:
                    # module wasn't previously loaded
                    continue

                if adapter.check_type(x):
                    self.type2adapter[t] = adapter
                    break
            else:
                raise RuntimeError(f"Serialization for type {t} not available.")
        adapter = self.type2adapter[t]
        result = adapter.to_dict(x)
        result[MAGIC_KEY] = adapter.typestring
        self.serialization_stack.remove(x_id)
        return result

    def _deserialize_object(self, x: dict) -> Any:
        assert isinstance(x, dict) and MAGIC_KEY in x

        typestring = x[MAGIC_KEY]
        convertor = self.adapters[typestring]

        if convertor is None:
            raise RuntimeError(f'Problems: adapter for {typestring} not found. Object: {x}')
        return convertor.from_dict(x)

    def serialize_helper(self, x):
        """
        Serialize helper should be passed as a hook to serialization to msgpack or downstream protocol
        """
        assert isinstance(x, LazyWrapper)
        if isinstance(x.obj, dict):
            return pack_dict(x.obj)
        else:
            return self._serialize_object(x.obj)

    def deserialize_helper(self, x):
        """
        Deserialize helper should be passed as a hook to serialization to msgpack or downstream protocol
        """
        if not isinstance(x, dict):
            return x
        elif MAGIC_KEY not in x:
            # usual dictionary
            return x
        else:
            return self._deserialize_object(x)


class AbstractAdapter:
    """
    Adapters should implement this interface.
    """
    package_name: Optional[str] = None
    typestring: str

    # adapter may introduce additional fields to configure save/load
    # if it is unavoidable

    def check_type(self, obj):
        raise NotImplementedError()

    def to_dict(self, obj) -> Dict[str, Any]:
        raise NotImplementedError()

    def from_dict(self, d: Dict[str, Any]):
        raise NotImplementedError()


def pack_dict(d):
    if all(isinstance(x, str) for x in d) and (MAGIC_KEY not in d):
        return {k: wrap(v) for k, v in d.items()}
    return {
        MAGIC_KEY: 'dict',
        'k': [wrap(k) for k in d.keys()],
        'v': [wrap(v) for v in d.values()],
    }


class DictAdapter(AbstractAdapter):
    typestring = 'dict'

    # this adapter is exceptional
    # to_dict can't be implemented, we use it only during unpacking.
    # For packing use pack_dict
    def check_type(self, obj):
        return False

    def to_dict(self, obj) -> Dict[str, Any]:
        raise NotImplementedError('This will never be called')

    def from_dict(self, d):
        k, v = d['k'], d['v']
        return dict(zip(k, v))
